//
// randomm3u.cpp : Randomizes an .m3u playlist.
//
// Hacked together by Jeff Lait, who absolves all responsibility
// for the state of this code or future use of this code.
//
// One would normally use perl or something for this,
// but I wanted to be *very* sure that the order is random.
//
// All net-new code is distributed as Public Domain.  This excludes
// the Mersenne Twister code which is bound by the Artistic License IIRC.
//
// Thanks go to Mads Haahr of www.random.org for making true random
// numbers no longer a computing fable.
//
// Thanks go to Jochen Brenner (jochen_brenner at gmx dot de),
// whose sample file for how to read from random.org was referenced
// for the corresponding code here.
//
// Thanks go to Takuji Nishimura and Makoto Matsumoto, whose
// most excellent Mersenne Twister is used for the offline
// component.
//
// To compile under Linux:
//   g++ -O3 -Wall randomm3u.cpp -o randomm3u
// To compiler under Windows:
//   Make a project with this as the sole file.  Turn off precompiled
//   headers.
//

#define VERSION "1.1"

#include <stdio.h>
#include <stdlib.h>
#include <time.h>
#include <errno.h>

// The beauty of standards is that we have so many to choose from.
#ifdef _WIN32
    #include <tchar.h>
    #include <winsock.h>
    #include <io.h>

    #define OurSleep(x) Sleep((x)*1000)
#else
    #include <string.h>
    #include <ctype.h>
    #include <unistd.h>
    #include <sys/types.h>
    #include <sys/socket.h>
    #include <netinet/in.h>
    #include <netdb.h>

    #define closesocket(s) close(s)
    #define OurSleep(x) sleep(x)
#endif

// Tracks what the user specified as the host, ie:
// www.random.org
const char *glbRandomHost = 0;

// This is the email which is added to the user agent so
// people can contact the user of the agent rather than the
// hapless person who compiled it.
const char *glbUserEmail = 0;

// How many bytes we have used from our random number generator.
int glbUsedRandomBytes = 0;

// Reads a full line from a file pointer, regardless of
// its length.
// The resulting pointer must be freed by the caller.
char *
readfullline(FILE *fp)
{
    char		 buf[8192];
    char		*line = 0;
    int			 linelen = 0, linesize = 0, buflen;
    char		*success;

    while (1)
    {
	success = fgets(buf, 8190, fp);

	// If we failed to read, return immediately with our
	// line.  If we failed before reading anything, this will
	// return 0, otherwise, the partial line.
	// (we are falsely assuming failure implies EOF.)
	if (!success)
	    return line;

	buflen = (int) strlen(buf);

	// Append buf to line.  Be generous for null terminators.
	while (linelen + buflen >= linesize-1)
	{
	    // Resize line.
	    char	*temp;
	    temp = line;

	    // A simple doubling heuristic.
	    linesize = (linesize + 100) * 2;

	    line = (char *) malloc(linesize);
	    if (temp)
	    {
		strcpy(line, temp);
		free(temp);
	    }
	    else
	    {
		line[0] = '\0';
	    }
	}

	// We now have enough room.
	strcat(line, buf);
	linelen += buflen;

	// Determine if we have a new line...
	// (We opened in text mode, so I hope only \n, but I'm
	// paranoid.  This raises the problem of the \n\r occuring
	// across a buffer break... *sigh*)
	if (strchr(buf, '\n') || strchr(buf, '\r'))
	{
	    // We have a new line!  Finished!
	    return line;
	}

	// Continue appending...
    }

    // Dead code path
    return line;
}

//
// Queries the buf backwards for a valid hex byte.
// Fatal errors on any problem.
//
unsigned char
getNibble(unsigned char *buf, int &cursize)
{
    cursize--;
    if (cursize < 0)
    {
	printf("Underflow on data from www.random.org!\n");
	exit(-5);
    }

    buf[cursize] = toupper(buf[cursize]);

    if (buf[cursize] >= '0' && buf[cursize] <= '9')
    {
	return buf[cursize] - '0';
    }

    if (buf[cursize] >= 'A' && buf[cursize] <= 'F')
    {
	return buf[cursize] - 'A' + 10;
    }

    // Bad character, silently eat.
    return getNibble(buf, cursize);
}

// Retrieves a byte from our stream.
unsigned char
getByte(unsigned char *buf, int &cursize)
{
    unsigned char result;

    result = getNibble(buf, cursize);
    result <<= 4;
    result |= getNibble(buf, cursize);
    return result;
}

// Retrieves an integer from our stream.
int
getInteger(unsigned char *buf, int &cursize)
{
    int	    result;

    result = getByte(buf, cursize);
    result <<= 8;
    result |= getByte(buf, cursize);
    result <<= 8;
    result |= getByte(buf, cursize);
    result <<= 8;
    result |= getByte(buf, cursize);

    return result;
}

//
// Empties a socket of all available data.
// Returns a pointer to a large buffer (which you
// must free) and updates the size pointer.
// The data is returned in reverse order!  This allows
// us to ignore the usual http header and just parse
// the juice random data...
//
unsigned char *
emptysocket(int sock, int &cursize)
{
    // Read out the result into a large buffer.
    // Because we are lazy and don't want to strip the
    // header, we read the data backwards.
    int		     count;
    unsigned char   *bigbuf = 0;
    char	     buffer[8192];

    while (1)
    {
	// Get the next block
	count = recv(sock, buffer, sizeof(buffer), 0);
	if (count <= 0)
	    break;

	// Append to our big block.
	unsigned char   *oldbuf;

	oldbuf = bigbuf;
	bigbuf = new unsigned char[cursize + count + 2];
	if (oldbuf)
	    memcpy(bigbuf, oldbuf, cursize);
	delete [] oldbuf;
	memcpy(&bigbuf[cursize], buffer, count);
	cursize += count;
	bigbuf[cursize] = '\0';
    } while (count > 0);

    return bigbuf;
}

//
// Connect to the globally defined web address.
// Send a request of the given URL.
// Return an allocated buffer (that you have to free) of the
// entire response.
// Fatal error if fails to connect in five tries.
// Returns the size of the buffer in bufsize.
//
unsigned char *
retrieveURL(const char *URL, int &bufsize, bool useragent)
{
    struct sockaddr_in server;
    struct hostent *host_info;
    int sock;
    char buffer[8192];

    // Create the socket
    sock = (int) socket(PF_INET, SOCK_STREAM, 0);
    if (sock < 0)
    {
	printf("Failed to create a socket!");
	exit(-6);
    }

    // Clear the server structure.
    memset(&server, 0, sizeof(server));

    // Find the actual host info.
    host_info = gethostbyname(glbRandomHost);

    if (!host_info)
    {
	// Failed to look up the server.
	printf("Could not resolve %s\n", glbRandomHost);
	exit(-7);
    }

    // Verify we do not have an overrun.
    if (((int)host_info->h_length) > (int) (sizeof(server) - (((char *)&server.sin_addr) - ((char *)&server))))
    {
	printf("Host entry overflow!\n");
	exit(-8);
    }

    // Copy the actual IP address over.
    // Why does one always deal with such ugly and opaque structures
    // when one deals with networking?  Is it some form of fetish?
    memcpy((char *) &server.sin_addr, 
		    *host_info->h_addr_list, 
		     host_info->h_length);

    // Finish filling out the server structure.
    server.sin_family = AF_INET;
    server.sin_port = htons(80);

    // Set up the connection.
    int	trials = 0;

    while (trials < 5)
    {
	int connectionerror;

	trials++;

	connectionerror = connect(sock,
				  (struct sockaddr *) &server,
				  sizeof(server));
	if (connectionerror < 0)
	{
	    printf("Failure to connect to %s\nError %d\n", glbRandomHost, connectionerror);
	    // We don't want to fatal error on a temporary connection
	    // failure...
	    if (trials < 5)
	    {
		printf("Backing off for 30 seconds...\n");
		OurSleep(30);
		printf("Trying again...\n");
	    }
	}
	else
	    break;
    }

    // Request the numbers...
    if (useragent)
	sprintf(buffer, "GET %s HTTP/1.0\nHost: %s\nUser-Agent: randomm3u " VERSION " (user email: %s)\n\n",
		    URL,
		    glbRandomHost,
		    glbUserEmail);
    else
	sprintf(buffer, "GET %s HTTP/1.0\nHost: %s\n\n",
		    URL,
		    glbRandomHost);
	
    send(sock, buffer, (int) strlen(buffer), 0);

    // Read out the result into a large buffer.
    unsigned char	*result = 0;

    result = emptysocket(sock, bufsize);

    // Close the connection.
    closesocket(sock);

    return result;
}


//
// Attempt to get random numbers from random.org.
// We get them in large blocks and filter appropriately.
// No seeding is required.
// Fatal error occurs if connection fails
//
int
get_randomorg()
{
#define CACHE_SIZE 256
    static int	    cachelist[CACHE_SIZE];
    static int	    cachepos = CACHE_SIZE;

    glbUsedRandomBytes += 4;

    if (cachepos >= CACHE_SIZE)
    {
	// Reload the cache.
	printf("Fetching new numbers from %s...\n", glbRandomHost);

	// Check the load on the server.
	while (1)
	{
	    unsigned char *checkbuf = 0;
	    int		   checksize = 0;

	    checkbuf = retrieveURL("/quota/?format=plain", checksize, true);

	    // Determine if negative, if so, we are out of bits!
	    int digitstart = checksize-1;
	    while (digitstart > 0 && isspace(checkbuf[digitstart]))
	    {
		checkbuf[digitstart] = '\0';
		digitstart--;
	    }

	    while (digitstart > 0 && (isdigit(checkbuf[digitstart]) ||
					checkbuf[digitstart] == '-'))
		digitstart--;
	    if (checkbuf[digitstart] != '-')
	    {
		printf("You have %s bits quota remaining.\n", &checkbuf[digitstart]);
		delete [] checkbuf;
		break;
	    }

	    printf("You are out of quota: %s bits.\nWaiting ten minutes to try again.\n", &checkbuf[digitstart]);
	    delete [] checkbuf;
	    int		tick;
	    for (tick = 0; tick < 10 * 6; tick++)
	    {
		OurSleep(10);
		printf("%d seconds...\n", (tick + 1) * 10);
	    }
	}

	// Read out the result into a large buffer.
	int		 cursize = 0;
	unsigned char	*bigbuf = 0;

	bigbuf = retrieveURL("/integers/?num=1024&min=0&max=255&col=1&base=16&format=plain&rnd=new", cursize, true);

	// We now have a huge buffer of size cursize.
	// There is some crappy header on the front of it we don't care about,
	// so we merely read the buffer backwards...
	for (cachepos = 0; cachepos < CACHE_SIZE; cachepos++)
	{
	    cachelist[cachepos] = getInteger(bigbuf, cursize);

	    // Clear out sign bit
	    cachelist[cachepos] &= 0x7FFFffff;
	}

	delete [] bigbuf;

	printf("Fresh numbers fetched.\n");

	// Reset the cache position.
	cachepos = 0;
    }

    // Return the next entry in the cache.
    return cachelist[cachepos++];
}

/*
 * Start of Mersenne Twister code
 * The code below has been highly abbreviated.  It is strongly
 * recommended you go to the original source if you wish to
 * use it in another project.
 */

/* 
   A C-program for MT19937, with initialization improved 2002/1/26.
   Coded by Takuji Nishimura and Makoto Matsumoto.

   Before using, initialize the state by using init_genrand(seed)  
   or init_by_array(init_key, key_length).

   Copyright (C) 1997 - 2002, Makoto Matsumoto and Takuji Nishimura,
   All rights reserved.                          

   Redistribution and use in source and binary forms, with or without
   modification, are permitted provided that the following conditions
   are met:

     1. Redistributions of source code must retain the above copyright
        notice, this list of conditions and the following disclaimer.

     2. Redistributions in binary form must reproduce the above copyright
        notice, this list of conditions and the following disclaimer in the
        documentation and/or other materials provided with the distribution.

     3. The names of its contributors may not be used to endorse or promote 
        products derived from this software without specific prior written 
        permission.

   THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS
   "AS IS" AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT
   LIMITED TO, THE IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR
   A PARTICULAR PURPOSE ARE DISCLAIMED.  IN NO EVENT SHALL THE COPYRIGHT OWNER OR
   CONTRIBUTORS BE LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL,
   EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT LIMITED TO,
   PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES; LOSS OF USE, DATA, OR
   PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND ON ANY THEORY OF
   LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT (INCLUDING
   NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE OF THIS
   SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.


   Any feedback is very welcome.
   http://www.math.sci.hiroshima-u.ac.jp/~m-mat/MT/emt.html
   email: m-mat @ math.sci.hiroshima-u.ac.jp (remove space)
*/

#include <stdio.h>

/* Period parameters */  
#define N 624
#define M 397
#define MATRIX_A 0x9908b0dfUL   /* constant vector a */
#define UPPER_MASK 0x80000000UL /* most significant w-r bits */
#define LOWER_MASK 0x7fffffffUL /* least significant r bits */

static unsigned long mt[N]; /* the array for the state vector  */
static int mti=N+1; /* mti==N+1 means mt[N] is not initialized */

/* initializes mt[N] with a seed */
void init_genrand(unsigned long s)
{
    mt[0]= s & 0xffffffffUL;
    for (mti=1; mti<N; mti++) {
        mt[mti] = 
	    (1812433253UL * (mt[mti-1] ^ (mt[mti-1] >> 30)) + mti); 
        /* See Knuth TAOCP Vol2. 3rd Ed. P.106 for multiplier. */
        /* In the previous versions, MSBs of the seed affect   */
        /* only MSBs of the array mt[].                        */
        /* 2002/01/09 modified by Makoto Matsumoto             */
        mt[mti] &= 0xffffffffUL;
        /* for >32 bit machines */
    }
}

/* initialize by an array with array-length */
/* init_key is the array for initializing keys */
/* key_length is its length */
/* slight change for C++, 2004/2/26 */
void init_by_array(unsigned long init_key[], int key_length)
{
    int i, j, k;
    init_genrand(19650218UL);
    i=1; j=0;
    k = (N>key_length ? N : key_length);
    for (; k; k--) {
        mt[i] = (mt[i] ^ ((mt[i-1] ^ (mt[i-1] >> 30)) * 1664525UL))
          + init_key[j] + j; /* non linear */
        mt[i] &= 0xffffffffUL; /* for WORDSIZE > 32 machines */
        i++; j++;
        if (i>=N) { mt[0] = mt[N-1]; i=1; }
        if (j>=key_length) j=0;
    }
    for (k=N-1; k; k--) {
        mt[i] = (mt[i] ^ ((mt[i-1] ^ (mt[i-1] >> 30)) * 1566083941UL))
          - i; /* non linear */
        mt[i] &= 0xffffffffUL; /* for WORDSIZE > 32 machines */
        i++;
        if (i>=N) { mt[0] = mt[N-1]; i=1; }
    }

    mt[0] = 0x80000000UL; /* MSB is 1; assuring non-zero initial array */ 
}

/* generates a random number on [0,0xffffffff]-interval */
unsigned long genrand_int32(void)
{
    glbUsedRandomBytes += 4;

    unsigned long y;
    static unsigned long mag01[2]={0x0UL, MATRIX_A};
    /* mag01[x] = x * MATRIX_A  for x=0,1 */

    if (mti >= N) { /* generate N words at one time */
        int kk;

        if (mti == N+1)   /* if init_genrand() has not been called, */
            init_genrand(5489UL); /* a default initial seed is used */

        for (kk=0;kk<N-M;kk++) {
            y = (mt[kk]&UPPER_MASK)|(mt[kk+1]&LOWER_MASK);
            mt[kk] = mt[kk+M] ^ (y >> 1) ^ mag01[y & 0x1UL];
        }
        for (;kk<N-1;kk++) {
            y = (mt[kk]&UPPER_MASK)|(mt[kk+1]&LOWER_MASK);
            mt[kk] = mt[kk+(M-N)] ^ (y >> 1) ^ mag01[y & 0x1UL];
        }
        y = (mt[N-1]&UPPER_MASK)|(mt[0]&LOWER_MASK);
        mt[N-1] = mt[M-1] ^ (y >> 1) ^ mag01[y & 0x1UL];

        mti = 0;
    }
  
    y = mt[mti++];

    /* Tempering */
    y ^= (y >> 11);
    y ^= (y << 7) & 0x9d2c5680UL;
    y ^= (y << 15) & 0xefc60000UL;
    y ^= (y >> 18);

    return y;
}

/* generates a random number on [0,0x7fffffff]-interval */
long genrand_int31(void)
{
    return (long)(genrand_int32()>>1);
}

/*
 * End of Mersenne Twister code
 */

//
// Returns a random number from 0..range inclusive.
//
// We require a source of unsigned random numbers.  We use 
// random.org if a host is specified, otherwise we use Mersenne Twister.
//
// We do not just do a modulus of the number, as that will result in
// biases.  Ie, say we have a random distribution 0..15.  If we modulus
// with 10, we'd get 0..5 being twice as likely as 6..9.  Of course,
// when the modulus range is small compared to the random range
// (ie, 100 vs 2,000,000,000) this difference is small.  But it would
// seem foolish in the extreme to go to all this work to get random
// numbers and then stop here!
//
// Rather than try anything complicated (like convert to float and then
// multiply to fill the range) which I'd likely screw up, I go for a
// dead simple approach.
//
// The source number is masked to lie within the first power of two
// bracket greater than the desired range.   This form of modulusing
// preserves uniformity.  We then test to see if we are in the range.
// If not, we try again.  Theoritically, we may have to wait forever
// for this to terminate :>  Practically, we have at least a 50%
// chance of getting a valid number, so we'll not have to wait too long.
//
int
true_rand(int range)
{
    if (!range)
	return 0;

    if (range < 0)
	return -true_rand(-range);

    // Find the next largest power of two.
    int	    mask = 1;
    while (mask <= range)
	mask <<= 1;

    // Find the bit mask.
    mask--;

    // Now repeatedly rand & mask until less than range.
    // This avoids any round off problems with large values of range
    int	    result;
    while (1)
    {
	if (glbRandomHost)
	    result = get_randomorg();
	else
	    result = genrand_int32();
	result &= mask;
	if (result <= range)
	    return result;
    }
}

#ifdef _WIN32
int _tmain(int argc, _TCHAR *argv[])
#else
int main(int argc, char *argv[])
#endif
{
    if (argc < 3)
    {
	printf("randomm3u version " VERSION "\n\n");
	printf("Usage: \n");
	printf("\t%s [infile.m3u] [outfile.m3u] [www.random.org youremail@domain.com]\n", argv[0]);
	return -1;
    }

    if (argc > 3)
    {
	printf("Will query %s for random numbers.\n", argv[3]);
	glbRandomHost = argv[3];
	if (argc < 5)
	{
	    printf("Missing email address!\nPlease provide so that the people at %s can contact you rather than blacklisting you if things go wrong.\n", glbRandomHost);
	    exit(-10);
	}
	glbUserEmail = argv[4];
    }

    if (glbRandomHost)
    {
	// The user requested an internet lookup of a host.
	// We thus need to initialize sockets on windows.
#ifdef _WIN32  
	short   	wVersionRequested;
	WSADATA 	wsaData;

	wVersionRequested = MAKEWORD (1, 1);
	
	// Request version 1.1 of sockets.
	// I have the sinking suspicion everyone requests 1.1 or 2.2
	// because they don't know which of the 1's represents the
	// major version number.  I wasted a few brain cells trying
	// to figure it out from the MSVC help...
	if (WSAStartup(wVersionRequested, &wsaData) != 0) 
	{
	    printf("Failed to init windows sockets\n");
	    return -4;
	}
#endif
    }

    // Open the source file.
    FILE	*fp;

    char  *header;
    char  *line;
    char **linelist = 0;
    char  *temp;

    int	    numlines = 0;
    int	    linelistsize = 0;
    int	    numentries;
    int	    i, j;

    fp = fopen(argv[1], "rt");

    if (!fp)
    {
	printf("Failed to open %s for reading.\n");
	return -2;
    }

    // Grab the header
    header = readfullline(fp);

    // Get all the lines.
    while ( (line = readfullline(fp)) )
    {
	if (numlines >= linelistsize)
	{
	    // Resize linelist:
	    char **temp = linelist;

	    // Double number of lines.
	    linelistsize = (linelistsize + 100) * 2;

	    linelist = new char *[linelistsize];
	    
	    // Copy over old lines.
	    if (temp)
	    {
		memcpy(linelist, temp, sizeof(char *) * numlines);
		delete [] temp;
	    }
	}
	linelist[numlines++] = line;
    }

    fclose(fp);

    printf("Found %d lines\n", numlines);

    // Now randomize.
    // We only need this for when not using a true
    // random number generator.
    init_genrand((unsigned long) time(0));

    // We shuffle pairs.
    numentries = numlines / 2;

    for (i = 0; i < numentries; i++)
    {
	// true_rand returns from 0..i inclusive
	j = true_rand(i);
	// Swap the two.
	if (i == j)
	    continue;
	temp = linelist[i*2];
	linelist[i*2] = linelist[j*2];
	linelist[j*2] = temp;
	temp = linelist[i*2+1];
	linelist[i*2+1] = linelist[j*2+1];
	linelist[j*2+1] = temp;
    }

    printf("Randomized.\n");

    // Write it out.
    fp = fopen(argv[2], "wt");
    if (!fp)
    {
	printf("Failure to open %s for write.\n", argv[2]);
	return -3;
    }

    // Write the header.
    fputs(header, fp);
    free(header);
    header = 0;

    for (i = 0; i < numlines; i++)
    {
	fputs(linelist[i], fp);
	free(linelist[i]);
    }
    delete [] linelist;
    linelist = 0;

    fclose(fp);

    printf("All done!\n");
    printf("%d bytes of entropy consumed.\n", glbUsedRandomBytes);

#ifdef _WIN32
    WSACleanup();
#endif

    return 0;
}

